# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from .texture import Texture, TextureIndex

__all__ = [
    "Texture",
    "TextureIndex",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from .face import FaceFilter
from .filter import TextureCombinedFilter, TextureConcatenatedFilter, TextureFilter
from .frame import FrameFilter
from .mipmap import MipmapFilter
from .resolution import ResolutionFilter
from .slice import SliceFilter

__all__ = [
    "TextureFilter",
    "TextureCombinedFilter",
    "TextureConcatenatedFilter",
    "MipmapFilter",
    "ResolutionFilter",
    "FaceFilter",
    "FrameFilter",
    "SliceFilter",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Literal

import numpy as np

from typing_extensions import deprecated

from no_vtf.core.image import ImageWithRawData

from .s3tc import decode_dxt1_rgb as _decode_dxt1_rgb
from .s3tc import decode_dxt1_rgba as _decode_dxt1_rgba
from .s3tc import decode_dxt3 as _decode_dxt3
from .s3tc import decode_dxt5 as _decode_dxt5


@deprecated("Use s3tc.decode_dxt1_rgb instead")
def decode_dxt1_rgb(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return _decode_dxt1_rgb(encoded_image, logical_width, logical_height)


@deprecated("Use s3tc.decode_dxt1_rgba instead")
def decode_dxt1_rgba(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return _decode_dxt1_rgba(encoded_image, logical_width, logical_height)


@deprecated("Use s3tc.decode_dxt3 instead")
def decode_dxt3(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return _decode_dxt3(encoded_image, logical_width, logical_height)


@deprecated("Use s3tc.decode_dxt5 instead")
def decode_dxt5(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return _decode_dxt5(encoded_image, logical_width, logical_height)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final, Literal, Optional, Union

from no_vtf.core.texture import Texture, TextureIndex
from no_vtf.typing import mypyc_attr

from .texture_index import TextureIndexFilter


@mypyc_attr(allow_interpreted_subclasses=True)
class SliceFilter(TextureIndexFilter[Texture]):
    def __init__(
        self,
        *,
        slices: slice,
        last: Union[Literal["original"], Literal["filtered"]] = "original",
        strict: bool = False,
    ) -> None:
        super().__init__(slice=slices, last=last, strict=strict)

        self.slices: Final = slices

    def select_texture_index(self, texture: Texture) -> Optional[TextureIndex]:
        return texture.slice



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from collections.abc import Callable
from typing import Final, Generic, Optional, TypeVar, cast

from no_vtf.typing import mypyc_attr

_A = TypeVar("_A")
_A_co = TypeVar("_A_co", covariant=True)
_B_co = TypeVar("_B_co", covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class Deferred(Generic[_A_co]):
    @classmethod
    def pure(cls, a: _A, /) -> Deferred[_A]:
        return Deferred(lambda: a)

    @classmethod
    def join(cls, a: Deferred[Deferred[_A_co]], /) -> Deferred[_A_co]:
        return Deferred(lambda: a()())

    def __init__(self, thunk: Callable[[], _A_co], /) -> None:
        self._thunk: Final = thunk
        self._result: Optional[_A_co] = None
        self._result_valid = False

    def __call__(self) -> _A_co:
        if not self._result_valid:
            self._result = self._thunk()
            self._result_valid = True

        return cast(_A_co, self._result)

    def map(self, f: Callable[[_A_co], _B_co], /) -> Deferred[_B_co]:
        return Deferred(lambda: f(self()))

    def apply(self, f: Deferred[Callable[[_A_co], _B_co]], /) -> Deferred[_B_co]:
        return Deferred(lambda: f()(self()))

    def bind(self, f: Callable[[_A_co], Deferred[_B_co]], /) -> Deferred[_B_co]:
        return Deferred(lambda: f(self())())



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from collections.abc import Sequence
from typing import Final, Literal, Optional, TypeVar, Union

from no_vtf.core.texture import Texture, TextureIndex
from no_vtf.typing import mypyc_attr

from .filter import TextureFilter

_T_contra = TypeVar("_T_contra", bound=Texture, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureIndexFilter(TextureFilter[_T_contra]):
    def __init__(
        self,
        *,
        slice: slice,  # noqa: A002
        last: Union[Literal["original"], Literal["filtered"]] = "original",
        strict: bool = False,
    ) -> None:
        self.slice: Final = slice
        self.last: Final = last
        self.strict: Final = strict

    @abstractmethod
    def select_texture_index(self, texture: _T_contra) -> Optional[TextureIndex]: ...

    def __call__(self, textures: Sequence[_T_contra]) -> Sequence[_T_contra]:
        if not textures:
            return []

        texture_index_counts = {
            texture_index.count
            for texture in textures
            if (texture_index := self.select_texture_index(texture))
        }
        if not texture_index_counts:
            return [] if self.strict else textures

        length: int
        if self.last == "original":
            if len(texture_index_counts) != 1:
                filter_type_prefix = type(self).__name__.removesuffix("Filter")
                raise ValueError(
                    f"{filter_type_prefix} count must be the same for all filtered textures"
                )

            length = next(iter(texture_index_counts))
        else:
            length = (
                max(
                    texture_index.index
                    for texture in textures
                    if (texture_index := self.select_texture_index(texture))
                )
                + 1
            )

        indices = range(*self.slice.indices(length))

        textures_filtered: list[_T_contra] = []
        for index in indices:
            textures_filtered.extend(
                texture
                for texture in textures
                if (
                    (texture_index := self.select_texture_index(texture))
                    and texture_index.index == index
                )
                or (not self.strict and not texture_index)
            )
        return textures_filtered



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Protocol, TypeVar

from no_vtf.core.image import AnyImageWithRawData

_I_co = TypeVar("_I_co", bound=AnyImageWithRawData, covariant=True)


class ImageDecoder(Protocol[_I_co]):
    def __call__(
        self, encoded_image: bytes, logical_width: int, logical_height: int, /
    ) -> _I_co: ...



#!/usr/bin/env python3

# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import subprocess
import sys

from typing import NoReturn


def main() -> NoReturn:
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "--quiet",
            "--no-input",
            "--disable-pip-version-check",
            "install",
            "--no-warn-script-location",
            "nox[uv]",
        ],
        check=True,
    )

    sys.exit(
        subprocess.run(
            [
                sys.executable,
                "-m",
                "nox",
                "--non-interactive",
                *sys.argv[1:],
            ]
        ).returncode
    )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools
import itertools
import sys

from collections.abc import Callable, Iterator, Sequence
from contextlib import AbstractContextManager
from typing import Optional, Protocol, cast

import alive_progress
import alive_progress.animations.bars
import alive_progress.animations.spinner_compiler
import alive_progress.animations.spinners
import alive_progress.animations.utils
import alive_progress.styles.internal
import alive_progress.utils.cells
import click

from typing_extensions import ParamSpec

from ._click import posix_tty_style

_P = ParamSpec("_P")


class AliveBar(Protocol):
    text: str

    def __call__(self, *, skipped: bool = False) -> None: ...


def alive_bar(
    total: Optional[int] = None, *, receipt: bool = True
) -> AbstractContextManager[AliveBar]:
    style = functools.partial(posix_tty_style, io=sys.stderr)

    classic = bar_factory(
        style("=", fg=127, bold=True),
        tip=style(">", fg=127, bold=True),
        background=" ",
        borders=(
            style("[", fg=172, bold=True),
            style("]", fg=172, bold=True),
        ),
        underflow=style("!", fg="red", bold=True),
        overflow=style("x", fg="red", bold=True),
    )

    brackets = bouncing_spinner_factory(
        style(">" * 10, fg=127, bold=True), style("<" * 10, fg=127, bold=True)
    )

    alive_progress.styles.internal.BARS["no_vtf"] = classic
    alive_progress.styles.internal.SPINNERS["no_vtf"] = brackets

    return cast(
        AbstractContextManager[AliveBar],
        alive_progress.alive_bar(
            total,
            length=40,
            spinner=None,
            bar="no_vtf",
            unknown="no_vtf",
            file=sys.stderr,
            enrich_print=False,
            receipt=receipt,
        ),
    )


def bouncing_spinner_factory(chars_1: str, chars_2: str, *, right: bool = True) -> object:
    scroll_1 = scrolling_spinner_factory(chars_1, right=right)
    scroll_2 = scrolling_spinner_factory(chars_2, right=not right)
    return alive_progress.animations.spinners.sequential_spinner_factory(scroll_1, scroll_2)


def scrolling_spinner_factory(chars: str, *, right: bool = True) -> object:
    num_cells = len(alive_progress.utils.cells.to_cells(click.unstyle(chars)))
    natural = num_cells * 2

    @alive_progress.animations.spinner_compiler.spinner_controller(
        natural=natural,
    )  # type: ignore[misc]
    def inner_spinner_factory(actual_length: Optional[int] = None) -> object:
        def frame_data() -> Iterator[str]:
            nonlocal actual_length
            actual_length = actual_length or natural

            start = 0
            stop = actual_length - num_cells + 1
            frame_iterator: Sequence[int] = list(range(start, stop))
            if not right:
                frame_iterator = list(reversed(frame_iterator))

            for i in frame_iterator:
                yield " " * i + chars + " " * (actual_length - i - num_cells)

        return (frame_data(),)

    return inner_spinner_factory


def bar_factory(  # noqa: C901
    base: str,
    *,
    tip: str,
    background: str,
    underflow: str,
    overflow: str,
    borders: Optional[tuple[str, str]] = None,
) -> object:
    @alive_progress.animations.bars.bar_controller  # type: ignore[misc]
    def inner_bar_factory(
        length: int, spinner_factory: Optional[Callable[[int], object]] = None
    ) -> object:
        @bordered(borders, "||")
        def draw_known(
            running: bool, percent: float
        ) -> tuple[tuple[str, ...], Optional[tuple[str]]]:
            percent = max(0, percent)

            base_length = round(percent * (length + 1))
            tip_length = 0

            if base_length > 0:
                if base_length <= length:
                    tip_length += 1

                base_length = min(length, base_length - 1)

            underflow_length = 0
            underflow_border = False
            if not running and percent < 1:
                if base_length + tip_length < length:
                    underflow_length = 1
                else:
                    underflow_border = True

            background_length = length - (base_length + tip_length + underflow_length)

            rendered_base = base_length * (base,)
            rendered_tip = tip_length * (tip,)
            rendered_underflow = underflow_length * (underflow,)
            rendered_background = background_length * (background,)

            right_border = None
            if percent > 1:
                right_border = (overflow,)
            if underflow_border:
                right_border = (underflow,)

            return (
                rendered_base + rendered_tip + rendered_underflow + rendered_background,
                right_border,
            )

        if not spinner_factory:
            return draw_known, True, False, None

        player = alive_progress.animations.utils.spinner_player(spinner_factory(length))

        @bordered(borders, "||")
        def draw_unknown(percent: float) -> tuple[tuple[str, ...], Optional[tuple[str]]]:
            return next(player), None

        return draw_known, True, False, draw_unknown

    return inner_bar_factory


def bordered(borders: Optional[Sequence[str]], default: Sequence[str]) -> Callable[
    [Callable[_P, tuple[tuple[str, ...], Optional[tuple[str]]]]],
    Callable[_P, tuple[str, ...]],
]:
    def wrapper(
        fn: Callable[_P, tuple[tuple[str, ...], Optional[tuple[str]]]]
    ) -> Callable[_P, tuple[str, ...]]:
        @functools.wraps(fn)
        def inner_bordered(*args: _P.args, **kwargs: _P.kwargs) -> tuple[str, ...]:
            content, right = fn(*args, **kwargs)
            return tuple(itertools.chain(left_border, content, right or right_border))

        return inner_bordered

    left_border = tuple((borders or default)[0:1] or default[0:1])
    right_border = tuple((borders or default)[1:2] or default[1:2])
    return wrapper



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

from typing import Literal

import numpy as np
import numpy.typing as npt

from no_vtf.core.image import ImageWithRawData
from no_vtf.functional import Deferred

from .ndarray import s3tc_bytes_to_ndarray


def decode_dxt1_rgb(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        rgba8888 = s3tc_bytes_to_ndarray(
            encoded_image,
            logical_width,
            logical_height,
            np.uint8,
            mode="RGBA",
            n=1,
            pixel_format="DXT1",
        )
        rgb888: npt.NDArray[np.uint8] = rgba8888[..., :3]
        return rgb888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_dxt1_rgba(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="RGBA",
        n=1,
        pixel_format="DXT1",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt3(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="RGBA",
        n=2,
        pixel_format="DXT3",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt5(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="RGBA",
        n=3,
        pixel_format="DXT5",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_bc4(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["l"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="L",
        n=4,
        pixel_format="bc4",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="l"
    )


def decode_bc5(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="RGB",
        n=5,
        pixel_format="bc5",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_ati2n(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        grx888 = s3tc_bytes_to_ndarray(
            encoded_image,
            logical_width,
            logical_height,
            np.uint8,
            mode="RGB",
            n=5,
            pixel_format="bc5",
        )
        rgx888 = grx888[..., (1, 0, 2)]
        return rgx888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_bc7(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        s3tc_bytes_to_ndarray,
        encoded_image,
        logical_width,
        logical_height,
        np.uint8,
        mode="RGBA",
        n=7,
        pixel_format="bc7",
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import functools
import inspect
import os
import pathlib
import re
import shlex
import shutil
import sys

from collections.abc import Sequence
from typing import IO, Any, Final, Optional, TextIO, TypeVar, Union

import click
import click.shell_completion
import click_option_group

_AnyStr = TypeVar("_AnyStr", bytes, str)
_T = TypeVar("_T")


class _EnableDynamicInheritanceMeta(type):
    def __getitem__(cls, bases: Union[type, Sequence[type]], /) -> _EnableDynamicInheritanceMeta:
        if isinstance(bases, type):
            bases = [bases]

        bases = (cls,) + tuple(bases)
        name = f"_{'_'.join(base.__name__ for base in bases)}"
        return type(cls)(name, bases, {})


class ShellCompletionOption(click.Option, metaclass=_EnableDynamicInheritanceMeta):
    _shells: Final = {
        "bash": [
            "{prog_name} {shell_completion_opt} >> ~/.bashrc",
            "{complete_var}=bash_source {console_script}",
            '[ ! -x {console_script} ] || eval "$({complete_source})"',
            """
            if [ -x {console_script} ]; then
                {complete_var}_CACHE="${{XDG_CACHE_HOME:-${{HOME}}/.cache}}/{complete_subdir}/{shell}"
                if [ ! -e "${complete_var}_CACHE" ] || \
[ {console_script} -nt "${complete_var}_CACHE" ]; then
                    mkdir --parents "$(dirname "${complete_var}_CACHE")"
                    {complete_source} >"${complete_var}_CACHE"
                fi
                source "${complete_var}_CACHE"
                unset {complete_var}_CACHE
            fi
            """,
        ],
        "zsh": [
            "{prog_name} {shell_completion_opt} >> ~/.zshrc",
            "{complete_var}=zsh_source {console_script}",
            '[ ! -x {console_script} ] || eval "$({complete_source})"',
            """
            if [ -x {console_script} ]; then
                {complete_var}_CACHE="${{XDG_CACHE_HOME:-${{HOME}}/.cache}}/{complete_subdir}/{shell}"
                if [ ! -e "${complete_var}_CACHE" ] || \
[ {console_script} -nt "${complete_var}_CACHE" ]; then
                    mkdir --parents "$(dirname "${complete_var}_CACHE")"
                    {complete_source} >"${complete_var}_CACHE"
                fi
                source "${complete_var}_CACHE"
                unset {complete_var}_CACHE
            fi
            """,
        ],
        "fish": [
            "{prog_name} {shell_completion_opt} > ~/.config/fish/completions/{prog_name}.fish",
            "{complete_var}=fish_source {console_script}",
            '[ ! -x {console_script} ] || eval "$({complete_source})"',
            """
            if [ -x {console_script} ]
                if [ -n "$XDG_CACHE_HOME" ]
                    set {complete_var}_CACHE "$XDG_CACHE_HOME"'/{complete_subdir}/{shell}'
                else
                    set {complete_var}_CACHE "$HOME"'/.cache/{complete_subdir}/{shell}'
                end
                if [ ! -e "${complete_var}_CACHE" ] || \
[ {console_script} -nt "${complete_var}_CACHE" ]
                    mkdir --parents "$(dirname "${complete_var}_CACHE")"
                    {complete_source} >"${complete_var}_CACHE"
                end
                source "${complete_var}_CACHE"
                set -e {complete_var}_CACHE
            end
            """,
        ],
    }

    _env_var_name_pattern: Final[re.Pattern[str]] = re.compile(
        r"(?![0-9])[a-z_0-9]+", re.ASCII | re.IGNORECASE
    )

    def __init__(
        self,
        param_decls: Optional[Sequence[str]] = None,
        *,
        help: Optional[str] = None,  # noqa: A002
        hidden: bool = False,
        cache: bool = True,
        complete_var: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        if not param_decls:
            param_decls = ["--shell-completion"]

        if not help:
            help = "Setup shell completion."  # noqa: A001

        hidden = hidden or not self._is_platform_supported()

        super().__init__(
            param_decls,
            help=help,
            hidden=hidden,
            type=str,
            metavar="[SHELL]",
            is_flag=False,
            flag_value="",
            expose_value=False,
            is_eager=True,
            callback=self._callback,
            **kwargs,
        )

        self.cache: Final = cache
        self.complete_var: Final = complete_var

    def _callback(
        self, ctx: click.Context, param: click.Parameter, shell: Optional[str], /
    ) -> None:
        if shell is None or ctx.resilient_parsing:
            return

        try:
            if not self._is_platform_supported():
                raise RuntimeError("Only the Linux platform is supported")

            strict = not sys.stdout.isatty()
            io = sys.stdout if strict else sys.stderr
            echo = functools.partial(click.echo, file=io)
            style = functools.partial(posix_tty_style, io=io)

            console_script_path = self._detect_console_script_path()
            prog_name = console_script_path.name
            complete_var = self._get_complete_var(prog_name)

            shell = shell or self._detect_shell(strict=strict)
            shell = self._check_supported_shell(shell, strict=strict)

            print_shells = [
                print_shell for print_shell in self._shells if shell in {print_shell, None}
            ]
            for shell in print_shells:
                shell_desc = self._shells[shell]

                append_command = shell_desc[0].format(
                    prog_name=shlex.quote(prog_name),
                    shell_completion_opt=self.opts[0],
                )
                complete_source = shell_desc[1].format(
                    complete_var=complete_var,
                    console_script=shlex.quote(str(console_script_path)),
                )

                if not self.cache:
                    complete_command = shell_desc[2].format(
                        console_script=shlex.quote(str(console_script_path)),
                        complete_source=complete_source,
                    )
                else:
                    complete_command = (
                        inspect.cleandoc(shell_desc[3])
                        .replace(4 * " ", "\t")
                        .format(
                            console_script=shlex.quote(str(console_script_path)),
                            complete_var=complete_var,
                            complete_subdir=complete_var.removeprefix("_").lower(),
                            shell=shell,  # noqa S604 false positive
                            complete_source=complete_source,
                        )
                    )

                if not strict:
                    echo(
                        os.fsencode(
                            f"In {shell}, execute "
                            + f"{style(append_command, bold=True)}"
                            + f" to append the following {'lines' if self.cache else 'line'}:"
                        )
                    )

                echo("")
                echo(os.fsencode(complete_command))

                if not strict:
                    echo("")

            if not strict:
                echo("Finally, restart the shell to load the completion support.")

        except Exception as ex:
            error(str(ex))
            ctx.exit(1)

        ctx.exit()

    def _is_platform_supported(self) -> bool:
        return sys.platform == "linux"

    def _detect_console_script_path(self) -> pathlib.Path:
        if override := os.environ.get("_CONSOLE_SCRIPT_PATH_OVERRIDE"):
            return pathlib.Path(override)

        # click/utils.py, _detect_program_name()
        if getattr(sys.modules["__main__"], "__package__", None):
            raise RuntimeError(
                "The program must be run as a console script command, not as a module"
            )

        cmd_path = shutil.which(sys.argv[0])
        if not cmd_path:
            raise RuntimeError(f"Cannot determine command path for {sys.argv[0]!r}")

        full_path = pathlib.Path(cmd_path).resolve(strict=True)
        if full_path.name.endswith(".py"):
            raise RuntimeError(
                "The program must be run as a console script command, not as a Python script"
            )

        return full_path

    def _get_complete_var(self, prog_name: str) -> str:
        complete_var = self.complete_var

        if not complete_var:
            # click/core.py, _main_shell_completion()
            complete_name = prog_name.replace("-", "_").replace(".", "_")
            complete_var = f"_{complete_name}_COMPLETE".upper()

            if not self._env_var_name_pattern.fullmatch(complete_var):
                raise RuntimeError(
                    "Only letters, digits, and underscores are allowed in the program name"
                )

        return complete_var

    def _detect_shell(self, *, strict: bool) -> Optional[str]:
        try:
            return pathlib.Path(f"/proc/{os.getppid()}/exe").resolve(strict=True).name
        except Exception as ex:
            message = f"Cannot detect shell: {str(ex)}"

            if not strict:
                error(f"{message}\n")
                return None
            else:
                raise RuntimeError(message) from ex

    def _check_supported_shell(self, shell: Optional[str], *, strict: bool) -> Optional[str]:
        if shell and shell in self._shells:
            return shell

        if shell:
            message = (
                f"The {sanitize(shell, unsafe_extras=' ')} shell is not supported. "
                + f"Supported shells: {', '.join(self._shells)}."
            )

            if not strict:
                warning(f"{message}\n")
            else:
                raise RuntimeError(message)

        return None


class HiddenFromShellCompletionOption(click.Option):
    @property
    def hidden(self) -> bool:
        # during shell completion there is no current Click context
        if not click.get_current_context(silent=True):
            return True

        return bool(vars(self).get("hidden"))

    @hidden.setter
    def hidden(self, value: bool, /) -> None:
        vars(self)["hidden"] = value


class HelpFormatter(click.HelpFormatter):
    def write_usage(self, prog: str, args: str = "", prefix: Optional[str] = None) -> None:
        prog = click.style(prog, fg=127, bold=True)
        args = click.style(args, bold=True)
        super().write_usage(prog, args, prefix)

    def write_heading(self, heading: str) -> None:
        heading = click.style(heading, underline=True)
        super().write_heading(heading)


class OptionGroup(click_option_group.OptionGroup):
    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if not help_record:
            return None

        name, help_ = help_record
        name = click.style(name, fg=172, bold=True)
        return name, help_


class GroupedOption(click_option_group.GroupedOption):
    _dim_pattern: Final[re.Pattern[str]] = re.compile(r"(?<=\s)\[[^\[\]\s]+\]", re.ASCII)

    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if help_record is None:
            return None

        def dim_repl(match: re.Match[str]) -> str:
            return click.style(match.group(), dim=True)

        opts, opt_help = help_record
        opt_help = self._dim_pattern.sub(dim_repl, opt_help)
        return opts, opt_help


class Slice(click.ParamType):
    name = "slice"

    def get_metavar(self, param: click.Parameter) -> str:
        return "INDEX|[START]:[STOP][:STEP]"

    def convert(
        self,
        value: Union[str, Sequence[slice], slice],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Sequence[slice]:
        if isinstance(value, slice):
            return [value]
        if not isinstance(value, str):
            return value

        return [self._convert_single(single, param, ctx) for single in value.split(",")]

    def _convert_single(
        self,
        value: str,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> slice:
        slice_indices = self._to_slice_indices(value, param, ctx)

        start = slice_indices[0]
        if len(slice_indices) == 1:
            if start is None:
                self.fail("Index is empty.", param, ctx)
            if start >= 0:
                return slice(start, start + 1)
            else:
                stop = start + 1 if start != -1 else None
                return slice(start, stop)

        stop = slice_indices[1]
        if len(slice_indices) == 2:
            return slice(start, stop)

        step = slice_indices[2]
        if len(slice_indices) == 3:
            if step == 0:
                self.fail("Slice step cannot be zero.", param, ctx)
            return slice(start, stop, step)

        self.fail(f"Too many values in {value!r}.", param, ctx)

    def _to_slice_indices(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> list[Optional[int]]:
        slice_indices: list[Optional[int]] = []
        for slice_index in value.split(":"):
            if not slice_index:
                slice_indices.append(None)
            else:
                try:
                    slice_indices.append(int(slice_index))
                except ValueError:
                    self.fail(f"{slice_index!r} is not a valid integer.", param, ctx)
        return slice_indices


def get_param_default(
    params: Sequence[click.core.Parameter], param_name: str, param_type: type[_T]
) -> _T:
    for param in params:
        if param.name == param_name:
            default = param.default
            if callable(default):
                default = default()

            if not isinstance(default, param_type):
                raise ValueError(f"{type(default) = } {param_type = }")

            return default

    raise ValueError(f"No such parameter: {param_name}")


def apply_param_default(
    params: Sequence[click.core.Parameter],
    param_name: str,
    param_type: type[_T],
    value: Optional[_T],
) -> _T:
    return value if value is not None else get_param_default(params, param_name, param_type)


def posix_tty_style(
    text: str,
    *,
    io: TextIO,
    fg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bold: Optional[bool] = None,
    dim: Optional[bool] = None,
    underline: Optional[bool] = None,
    overline: Optional[bool] = None,
    italic: Optional[bool] = None,
    blink: Optional[bool] = None,
    reverse: Optional[bool] = None,
    strikethrough: Optional[bool] = None,
    reset: bool = True,
) -> str:
    if os.name == "posix" and io.isatty():
        return click.style(
            text,
            fg=fg,
            bg=bg,
            bold=bold,
            dim=dim,
            underline=underline,
            overline=overline,
            italic=italic,
            blink=blink,
            reverse=reverse,
            strikethrough=strikethrough,
            reset=reset,
        )

    return text


def echo(
    message: str,
    file: Optional[IO[_AnyStr]] = None,
    nl: bool = True,
    err: bool = False,
    color: Optional[bool] = None,
) -> None:
    message = click.style("", reset=True) + message
    click.echo(message, file, nl, err, color)


def warning(
    message: str,
    *,
    item: Optional[str] = None,
) -> None:
    message = (
        click.style("Warning", fg="yellow")
        + (" while processing " + click.style(item, bold=True) if item else "")
        + ": "
        + message
    )
    echo(message, file=sys.stderr)


def error(
    message: str,
    *,
    item: Optional[str] = None,
) -> None:
    message = (
        click.style("Error", fg="red")
        + (" while processing " + click.style(item, bold=True) if item else "")
        + ": "
        + message
    )
    echo(message, file=sys.stderr)


def sanitize(string: str, /, *, unsafe_extras: Optional[Sequence[str]] = None) -> str:
    unsafe_extras = unsafe_extras or []

    if any(unsafe in string for unsafe in ["'", '"', *unsafe_extras]):
        return repr(string)

    if repr(string)[1:-1] != string:
        return repr(string)

    return string



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Literal, Optional, TypeVar

import numpy as np
import numpy.typing as npt
import PIL.Image

from typing_extensions import deprecated

ByteOrder = Literal["S", "<", ">", "=", "|", "L", "B", "N", "I"]

_S = TypeVar("_S", bound=np.generic)


def raw_bytes_to_ndarray(
    image: bytes,
    width: int,
    height: int,
    channel_order: Sequence[int],
    scalar_type: type[_S],
    byte_order: Optional[ByteOrder] = None,
) -> npt.NDArray[_S]:
    num_channels = len(channel_order)
    shape = (height, width, num_channels)

    dtype = np.dtype(scalar_type)
    if byte_order is not None:
        dtype = dtype.newbyteorder(byte_order)

    ndarray: npt.NDArray[_S] = np.ndarray(shape=shape, dtype=dtype, buffer=image)
    ndarray = ndarray[..., channel_order]
    return ndarray.copy()


def s3tc_bytes_to_ndarray(
    encoded_image: bytes,
    logical_width: int,
    logical_height: int,
    scalar_type: type[_S],
    *,
    mode: str,
    n: int,
    pixel_format: str,
) -> npt.NDArray[_S]:
    # reference for "n" and "pixel_format": Pillow/src/PIL/DdsImagePlugin.py
    pil_image = PIL.Image.frombytes(  # pyright: ignore [reportUnknownMemberType]
        mode, (logical_width, logical_height), encoded_image, "bcn", n, pixel_format
    )

    ndarray: npt.NDArray[_S] = np.array(pil_image, dtype=scalar_type)
    ndarray = ndarray[(...,) + tuple([np.newaxis] * (3 - ndarray.ndim))]
    return ndarray.copy()


@deprecated("Use raw_bytes_to_ndarray instead")
def image_bytes_to_ndarray(  # type: ignore[misc]
    image: bytes,
    width: int,
    height: int,
    channel_order: Sequence[int],
    scalar_type: type[_S],
    byte_order: Optional[ByteOrder] = None,
) -> npt.NDArray[_S]:
    return raw_bytes_to_ndarray(image, width, height, channel_order, scalar_type, byte_order)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from typing import TypeAlias, TypeVar

from no_vtf.core.image import AnyImageWithRawData
from no_vtf.typing import mypyc_attr

from .file import FileIO, FileIOReadbackOptions, FileIOWriteOptions
from .io import IOReadbackOptions, IOWriteOptions


@mypyc_attr(allow_interpreted_subclasses=True)
class RawIOWriteOptions(FileIOWriteOptions):
    pass


@mypyc_attr(allow_interpreted_subclasses=True)
class RawIOReadbackOptions(FileIOReadbackOptions):
    pass


_T_contra = TypeVar("_T_contra", bound=pathlib.Path, contravariant=True)
_I_contra = TypeVar("_I_contra", bound=AnyImageWithRawData, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class RawIO(FileIO[_T_contra, _I_contra]):
    def write_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOWriteOptions = IOWriteOptions(),
    ) -> bool:
        if not super().write_file(path, *images, options=options):
            return False

        with path.open("wb") as file:
            for image in images:
                data = image.raw
                file.write(data)

        return True

    def readback_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOReadbackOptions = IOReadbackOptions(),
    ) -> bool:
        if not super().readback_file(path, *images, options=options):
            return False

        with path.open("rb") as file:
            for image in images:
                data = image.raw

                read_data = file.read(len(data))
                if data != read_data:
                    raise ValueError(f"{path!r}: Data differs from what is in the file")

            if file.read():
                raise ValueError(f"{path!r}: Data differs from what is in the file")

        return True


AnyRawIO: TypeAlias = RawIO[pathlib.Path, AnyImageWithRawData]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from .fp_precision_modifier import FPPrecisionModifier
from .hdr_to_ldr_modifier import HdrToLdrModifier
from .modifier import ImageModifier

__all__ = [
    "ImageModifier",
    "FPPrecisionModifier",
    "HdrToLdrModifier",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Literal, Optional

import numpy as np
import numpy.typing as npt

from typing_extensions import deprecated

from no_vtf.core.image import ImageWithRawData
from no_vtf.core.image.decoder.raw import (
    decode_bgr888,
    decode_bgra8888,
    decode_rgb888,
    decode_rgba16161616_le,
)
from no_vtf.functional import Deferred


def decode_rgb888_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return _decode_bluescreen(decode_rgb888(encoded_image, width, height))


def decode_bgr888_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return _decode_bluescreen(decode_bgr888(encoded_image, width, height))


def _decode_bluescreen(
    image: ImageWithRawData[np.uint8, Literal["rgb"]]
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = image.data.map(_decode_bluescreen_data)
    return ImageWithRawData(raw=image.raw, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def _decode_bluescreen_data(rgb888: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    is_opaque: npt.NDArray[np.bool_] = rgb888 != (0, 0, 255)

    is_opaque_any = is_opaque.any(axis=2)
    assert not isinstance(is_opaque_any, np.bool)  # noqa S101 type narrowing

    is_opaque_any = is_opaque_any[..., np.newaxis]

    rgb888 *= is_opaque_any
    a8: npt.NDArray[np.uint8] = np.multiply(is_opaque_any, 255, dtype=np.uint8)

    rgba8888: npt.NDArray[np.uint8] = np.dstack((rgb888, a8))
    return rgba8888


def decode_bgra8888_hdr(
    encoded_image: bytes, width: int, height: int, *, overbright_factor: Optional[float]
) -> ImageWithRawData[np.float32, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.float32]:
        nonlocal overbright_factor
        if overbright_factor is None:
            overbright_factor = 16

        rgba8888 = decode_bgra8888(encoded_image, width, height).data()

        rgba32323232f: npt.NDArray[np.float32] = rgba8888.astype(np.float32) / 255.0
        rgba32323232f[:, :, [0, 1, 2]] *= rgba32323232f[:, :, [3]] * overbright_factor

        rgb323232f: npt.NDArray[np.float32] = rgba32323232f[..., :3]
        return rgb323232f

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.float32), channels="rgb"
    )


def decode_rgba16161616_le_hdr(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float32, Literal["rgba"]]:
    def thunk() -> npt.NDArray[np.float32]:
        rgba16161616 = decode_rgba16161616_le(encoded_image, width, height).data()
        # convert 4.12 fixed point stored as integer into floating point
        rgba32323232f: npt.NDArray[np.float32] = rgba16161616.astype(np.float32) / (1 << 12)
        return rgba32323232f

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.float32), channels="rgba"
    )


@deprecated("Use decode_rgb888_bluescreen instead")
def decode_rgb_uint8_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_rgb888_bluescreen(encoded_image, width, height)


@deprecated("Use decode_bgr888_bluescreen instead")
def decode_bgr_uint8_bluescreen(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_bgr888_bluescreen(encoded_image, width, height)


@deprecated("Use decode_bgra8888_hdr instead")
def decode_bgra_uint8_hdr(
    encoded_image: bytes, width: int, height: int, *, overbright_factor: Optional[float]
) -> ImageWithRawData[np.float32, Literal["rgb"]]:
    return decode_bgra8888_hdr(encoded_image, width, height, overbright_factor=overbright_factor)


@deprecated("Use decode_rgba16161616_le_hdr instead")
def decode_rgba_uint16_le_hdr(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float32, Literal["rgba"]]:
    return decode_rgba16161616_le_hdr(encoded_image, width, height)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Literal

import numpy as np
import numpy.typing as npt

from typing_extensions import deprecated

from no_vtf.core.image import ImageWithRawData
from no_vtf.functional import Deferred

from .ndarray import raw_bytes_to_ndarray


def decode_rgb888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_rgba8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_argb8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (1, 2, 3, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_bgr888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(lambda: raw_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_bgra8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_bgrx8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0, 3), np.uint8)
    ).map(lambda data: data[..., :3])
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_abgr8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (3, 2, 1, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_rgba16161616_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba16161616_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba16161616f_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_rgba16161616f_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_l8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["l"]]:
    data = Deferred(lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="l")


def decode_a8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["a"]]:
    data = Deferred(lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="a")


def decode_la88(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["la"]]:
    la88 = Deferred(lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=la88, dtype=np.dtype(np.uint8), channels="la")


def decode_uv88(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        rg88 = raw_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
        b8: npt.NDArray[np.uint8] = np.zeros(rg88.shape[:-1], dtype=np.uint8)
        rgb888: npt.NDArray[np.uint8] = np.dstack((rg88, b8))
        return rgb888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_uvwq8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_uvlx8888(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(
        lambda: raw_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint8)
    ).map(lambda data: data[..., :3])
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_bgra4444(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        bgra4444 = raw_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)

        b8 = np.bitwise_and(bgra4444[..., [0]], 0x0F)
        b8 = np.bitwise_or(b8, np.left_shift(b8, 4))

        g8 = np.bitwise_and(bgra4444[..., [0]], 0xF0)
        g8 = np.bitwise_or(g8, np.right_shift(g8, 4))

        r8 = np.bitwise_and(bgra4444[..., [1]], 0x0F)
        r8 = np.bitwise_or(r8, np.left_shift(r8, 4))

        a8 = np.bitwise_and(bgra4444[..., [1]], 0xF0)
        a8 = np.bitwise_or(a8, np.right_shift(a8, 4))

        rgba8888: npt.NDArray[np.uint8] = np.dstack((r8, g8, b8, a8))
        return rgba8888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_rgb565(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        rgb565 = raw_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
        rgb565_uint16 = rgb565.view(np.dtype(np.uint16).newbyteorder("<"))[..., 0]

        r5 = np.bitwise_and(np.right_shift(rgb565_uint16, 0x0), 0x1F).astype(np.uint8)
        g6 = np.bitwise_and(np.right_shift(rgb565_uint16, 0x5), 0x3F).astype(np.uint8)
        b5 = np.bitwise_and(np.right_shift(rgb565_uint16, 0xB), 0x1F).astype(np.uint8)

        r8 = np.round(r5.astype(np.float64) / 31 * 255).astype(np.uint8)
        g8 = np.round(g6.astype(np.float64) / 63 * 255).astype(np.uint8)
        b8 = np.round(b5.astype(np.float64) / 31 * 255).astype(np.uint8)

        rgb888 = np.dstack((r8, g8, b8))
        return rgb888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_bgr565(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        bgr888 = decode_rgb565(encoded_image, width, height).data()
        rgb888 = bgr888[..., (2, 1, 0)]
        return rgb888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_bgra5551(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        bgra5551 = raw_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
        bgra5551_uint16 = bgra5551.view(np.dtype(np.uint16).newbyteorder("<"))[..., 0]

        b5 = np.bitwise_and(np.right_shift(bgra5551_uint16, 0x0), 0x1F).astype(np.uint8)
        g5 = np.bitwise_and(np.right_shift(bgra5551_uint16, 0x5), 0x1F).astype(np.uint8)
        r5 = np.bitwise_and(np.right_shift(bgra5551_uint16, 0xA), 0x1F).astype(np.uint8)
        a1 = np.bitwise_and(np.right_shift(bgra5551_uint16, 0xF), 0x01).astype(bool)

        rgb555 = np.dstack((r5, g5, b5))
        rgb888 = np.round(rgb555.astype(np.float64) / 31 * 255).astype(np.uint8)

        a8 = np.multiply(a1, 255, dtype=np.uint8)

        rgba888 = np.dstack((rgb888, a8))
        return rgba888

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_bgrx5551(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(decode_bgra5551(encoded_image, width, height).data).map(
        lambda data: data[..., :3]
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


@deprecated("Use decode_rgb888 instead")
def decode_rgb_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return decode_rgb888(encoded_image, width, height)


@deprecated("Use decode_rgba8888 instead")
def decode_rgba_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_rgba8888(encoded_image, width, height)


@deprecated("Use decode_argb8888 instead")
def decode_argb_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_argb8888(encoded_image, width, height)


@deprecated("Use decode_bgr888 instead")
def decode_bgr_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return decode_bgr888(encoded_image, width, height)


@deprecated("Use decode_bgra8888 instead")
def decode_bgra_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_bgra8888(encoded_image, width, height)


@deprecated("Use decode_bgrx8888 instead")
def decode_bgrx_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return decode_bgrx8888(encoded_image, width, height)


@deprecated("Use decode_abgr8888 instead")
def decode_abgr_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_abgr8888(encoded_image, width, height)


@deprecated("Use decode_rgba16161616_be instead")
def decode_rgba_uint16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    return decode_rgba16161616_be(encoded_image, width, height)


@deprecated("Use decode_rgba16161616_le instead")
def decode_rgba_uint16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    return decode_rgba16161616_le(encoded_image, width, height)


@deprecated("Use decode_rgba16161616f_be instead")
def decode_rgba_float16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    return decode_rgba16161616f_be(encoded_image, width, height)


@deprecated("Use decode_rgba16161616f_le instead")
def decode_rgba_float16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    return decode_rgba16161616f_le(encoded_image, width, height)


@deprecated("Use decode_l8 instead")
def decode_l_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["l"]]:
    return decode_l8(encoded_image, width, height)


@deprecated("Use decode_a8 instead")
def decode_a_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["a"]]:
    return decode_a8(encoded_image, width, height)


@deprecated("Use decode_la88 instead")
def decode_la_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["la"]]:
    return decode_la88(encoded_image, width, height)


@deprecated("Use decode_uv88 instead")
def decode_uv_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return decode_uv88(encoded_image, width, height)


@deprecated("Use decode_uvwq8888 instead")
def decode_uvwq_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_uvwq8888(encoded_image, width, height)


@deprecated("Use decode_uvlx8888 instead")
def decode_uvlx_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    return decode_uvlx8888(encoded_image, width, height)


@deprecated("Use decode_bgra4444 instead")
def decode_bgra_uint4_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    return decode_bgra4444(encoded_image, width, height)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import contextlib
import fnmatch
import functools
import itertools
import json
import logging
import os
import pathlib
import re
import shutil
import subprocess
import sys
import sysconfig
import tempfile
import threading

from abc import abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from contextlib import AbstractContextManager, contextmanager
from dataclasses import dataclass
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    ClassVar,
    Concatenate,
    Final,
    Literal,
    Optional,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
    final,
    overload,
)

import nox
import nox.command
import nox.logger
import nox.popen

if sys.platform == "linux":
    import fcntl
    import pty
    import termios
    import tty

if sys.version_info >= (3, 11):
    from typing import Self, Unpack
elif TYPE_CHECKING:
    from typing_extensions import Self, Unpack

StrPath: TypeAlias = Union[str, os.PathLike[str]]

_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)

nox.needs_version = ">= 2024.4.15"

nox.options.default_venv_backend = "venv"
if sys.platform == "linux":
    nox.options.default_venv_backend = "uv|" + nox.options.default_venv_backend

nox.options.error_on_external_run = True
nox.options.error_on_missing_interpreters = True
nox.options.sessions = ["lint"]


def tag_cachedir(dir: pathlib.Path) -> None:  # noqa: A002
    dir.mkdir(parents=True, exist_ok=True)
    pathlib.Path(dir, "CACHEDIR.TAG").write_bytes(b"Signature: 8a477f597d28d172789f06886806bc55\n")


if __file__:
    tag_cachedir(pathlib.Path(__file__).parent / ".nox")


@nox.session
def lint(session: nox.Session) -> None:
    session.install("black[colorama] >= 24.10.0, < 25")
    session.install("flake8 >= 7.1.1, < 8")
    session.install("flake8-bandit >= 4.1.1, < 5")
    session.install("flake8-builtins >= 2.5.0, < 3")
    session.install("flake8-deprecated >= 2.2.1, < 3")
    session.install("flake8-pep585 >= 0.1.7, < 1")
    session.install("isort[colors] >= 5.13.2, < 6")
    session.install("mypy[faster-cache] >= 1.13.0, < 2")
    session.install("nox >= 2024.10.9, < 2025")
    session.install("pep8-naming >= 0.14.1, < 1")
    session.install("pyright >= 1.1.390, < 2")
    session.install("reuse >= 5.0.2, < 6")
    session.install("shellcheck-py >= 0.10.0.1, < 1")
    session.install("shfmt-py >= 3.7.0.1, < 4")

    session.install("types-Pillow >= 10.2.0.20240822, < 11")

    session.install(".")

    posargs_paths = session.posargs

    fix = False
    if posargs_paths and posargs_paths[0] == "--fix":
        posargs_paths = posargs_paths[1:]
        fix = True

    posargs_paths = [str(pathlib.Path(session.invoked_from, path)) for path in posargs_paths]

    default_paths = ["no_vtf", "noxfile.py", "builds", "ksy"]

    paths = [pathlib.Path(path) for path in (posargs_paths or default_paths)]
    py_paths = [path for path in paths if path.is_dir() or path.name.endswith(".py")]
    sh_paths = list(
        itertools.chain.from_iterable(
            (
                path.rglob("*.sh*")
                if path.is_dir()
                else [path] if fnmatch.fnmatch(path.name, "*.sh*") else []
            )
            for path in paths
        )
    )

    shfmt = ["shfmt", "--simplify", "--func-next-line"]

    if not fix:
        if py_paths:
            session.run(
                "mypy",
                "--pretty",
                "--show-error-context",
                "--explicit-package-bases",
                "--",
                *py_paths,
            )
            session.run("pyright", "--warnings", *py_paths)
            session.run("flake8", "--", *py_paths, silent=True)
            session.run("isort", "--check", "--diff", "--", *py_paths)
            session.run("black", "--check", "--diff", "--", *py_paths)

        if sh_paths:
            session.run(
                "shellcheck",
                "--norc",
                "--external-sources",
                "--severity=style",
                "--enable=all",
                "--exclude=" + ",".join(["SC2016", "SC2032", "SC2033", "SC2250", "SC2292"]),
                "--format=gcc",
                "--",
                *sh_paths,
            )
            session.run(*shfmt, "--diff", "--", *sh_paths)

        session.run("reuse", "lint", silent=True)
    else:
        if py_paths:
            session.run("isort", "--", *py_paths)
            session.run("black", "--", *py_paths)

        if sh_paths:
            session.run(*shfmt, "--write", "--", *sh_paths)


def coverage(session: nox.Session) -> None:
    session.install("coverage[toml] >= 7.6.9, < 8")
    session.install("tomli >= 2.2.1, < 3")

    # to measure coverage through a console script, the package has to be installed in editable mode
    session.install("-e", ".")

    nox_session_install_only_end(session)

    for coverage_data_file in pathlib.Path().glob(".coverage*"):
        coverage_data_file.unlink()

    coverage_run = ["no_vtf"]
    runner = CoverageTestRunner(coverage_run=coverage_run, paths=session.bin_paths, env=session.env)
    with TestSuite.with_archived_samples(
        runner=runner, session=session, test_classes=get_test_classes_from_env()
    ) as test_suite:
        test_suite.write()
        test_suite.readback()

    coverage = ["python", "-m", "coverage"]
    session.run(*coverage, "combine", "--quiet", "--append")
    session.run(*coverage, "html")

    if not session.posargs:
        return

    archive_name = str(pathlib.Path(session.invoked_from, session.posargs[0]))
    make_archive(archive_name, base_dir="htmlcov")


if os.name == "posix":
    coverage = nox.session(coverage)


@nox.session
def package(session: nox.Session) -> None:
    session.install("build >= 1.2.2.post1, < 2")

    nox_session_install_only_end(session)

    path_dist = pathlib.Path("dist")
    if path_dist.is_dir():
        dist_files = [path for path in path_dist.iterdir() if path.is_file()]
        for dist_file in dist_files:
            dist_file.unlink()

    session.run("python", "-m", "build", silent=True)

    path_sdist = next(path_dist.glob("*.tar.gz"))
    path_wheel = next(path_dist.glob("*.whl"))

    nox_session_run_pip(session.run, "install", "--force-reinstall", str(path_wheel))

    executable = ["no_vtf"]
    session.run(*executable, "--version")

    runner = TestRunner(cmd=executable, env=session.env, paths=session.bin_paths)
    with TestSuite.with_archived_samples(
        runner=runner, session=session, test_classes=get_test_classes_from_env()
    ) as test_suite:
        test_suite.run([], {})

    if len(session.posargs) >= 1:
        shutil.copy2(path_sdist, pathlib.Path(session.invoked_from, session.posargs[0]))

    if len(session.posargs) >= 2:
        shutil.copy2(path_wheel, pathlib.Path(session.invoked_from, session.posargs[1]))


@nox.session(reuse_venv=False)
def freeze(session: nox.Session) -> None:  # noqa: C901
    build_dir = pathlib.Path("build")
    root_dir = pathlib.Path("dist")
    base_dir = pathlib.Path("no_vtf")
    dst_dir = root_dir / base_dir

    session.install("pyinstaller >= 6.11.1, < 7")

    nox_session_install_only_end(session)

    # installing in editable mode will ensure the _version.py file is generated by setuptools_scm
    # run even with the --no-install flag
    nox_session_run_pip(session.run, "install", "--force-reinstall", "-e", ".")

    if int(os.environ.get("BUILD_NUMPY_FROM_SOURCE", 0)):
        packages = json.loads(
            (
                nox_session_run_pip(
                    session.run,
                    "list",
                    *["--format", "json"],
                    silent=True,
                )
                or "[]"
            )
            .strip()
            .split("\n")[-1]
        )
        numpy = next(package for package in packages if package["name"] == "numpy")

        nox_session_run_pip(
            session.run,
            *["install", "--force-reinstall"],
            *["--no-binary", "numpy"],
            f"numpy == {numpy['version']}",
        )

    tag_cachedir(build_dir)

    session.run("imageio_download_bin", "--package-dir", silent=True)

    pyinstaller_args: list[str] = []

    pyinstaller_args.extend(["--log-level", "WARN"])
    pyinstaller_args.extend(["--name", "no_vtf"])
    pyinstaller_args.extend(["--icon", "resources/pyinstaller/empty.ico"])

    # https://github.com/rsalmei/alive-progress/issues/123
    pyinstaller_args.extend(["--collect-data", "grapheme"])

    # bundle plugin binary dependencies downloaded by imageio_download_bin
    pyinstaller_args.extend(["--collect-binaries", "imageio"])

    pyinstaller_args.append("--noconfirm")
    pyinstaller_args.append("--clean")

    session.run("pyinstaller", *pyinstaller_args, "no_vtf/__main__.py")

    shutil.rmtree(  # pyright: ignore [reportDeprecated]
        dst_dir / "_internal/grapheme/cython", ignore_errors=True
    )

    if os.name == "posix":
        for shared_library in dst_dir.rglob("*.so*"):
            nox_command_run(["strip", "--strip-all", "--", str(shared_library)])

    executable = str(dst_dir / "no_vtf")
    if os.name == "nt":
        executable += ".exe"

    paths = ["."]
    nox_command_run([executable, "--version"], paths=paths, external=True)

    data_subdirs = ["common", os.name]
    data_dirs = [pathlib.Path("resources/pyinstaller") / subdir for subdir in data_subdirs]
    for data_dir in data_dirs:
        if not data_dir.exists():
            continue

        shutil.copytree(data_dir, dst_dir, copy_function=shutil.copy2, dirs_exist_ok=True)

    readme = "README.md"
    shutil.copy2(readme, dst_dir)
    shutil.copy2(f"{readme}.license", dst_dir)

    python_license_dir: pathlib.Path
    if os.name == "nt":
        python_license_dir = pathlib.Path(sys.base_prefix)
    else:
        python_license_dir = pathlib.Path(sysconfig.get_paths("posix_prefix")["stdlib"])

    python_license_file = python_license_dir / "LICENSE.txt"

    licenses_dst_dir = dst_dir / "LICENSES"
    licenses_dst_dir.mkdir(exist_ok=True)

    shutil.copy2(python_license_file, licenses_dst_dir / "LicenseRef-Python.txt")

    if os.name == "nt":
        spdx_license_ids: Sequence[str] = []
        for spdx_license_id in spdx_license_ids:
            shutil.copy2(f"LICENSES/LicenseRef-{spdx_license_id}.txt", licenses_dst_dir)

    nox_session_run_pip(session.run, "install", "--upgrade", "pip")

    nox_session_run_pip(session.run, "install", "pip-licenses >= 5.0.0, < 6")
    nox_session_run_pip(session.run, "install", "pipdeptree >= 2.24.0, < 3")
    nox_session_run_pip(session.run, "install", "reuse >= 5.0.2, < 6")

    dependencies = json.loads(
        session.run(
            "pipdeptree",
            "--warn=fail",
            "--json",
            "--packages",
            ",".join(
                [
                    "no-vtf",
                    "pyinstaller",
                ]
            ),
            silent=True,
        )
        or ""
    )

    session.run(
        "pip-licenses",
        "--from=all",
        "--ignore-packages=pyinstaller-hooks-contrib",
        "--with-authors",
        "--with-urls",
        "--format=plain-vertical",
        f"--output-file={licenses_dst_dir / 'LicenseRef-no-vtf.txt'}",
        "--packages",
        *(dependency["package"]["key"] for dependency in dependencies),
        silent=True,
    )

    lorem_ipsum_path = dst_dir / "_internal/setuptools/_vendor/jaraco/text/Lorem ipsum.txt"
    if lorem_ipsum_path.exists():
        with open(lorem_ipsum_path, "wb"):
            pass

    reuse = [
        "reuse",
        f"--root={dst_dir}",
    ]
    session.run(*reuse, "download", "--all", silent=True)
    session.run(*reuse, "lint", silent=True)

    runner = TestRunner(cmd=[executable], env=session.env, paths=paths, external=True)
    with TestSuite.with_archived_samples(
        runner=runner, session=session, test_classes=get_test_classes_from_env()
    ) as test_suite:
        test_suite.run([], {})

    if not session.posargs:
        return

    if len(session.posargs) >= 2:
        path_sdist = pathlib.Path(session.invoked_from, session.posargs[1])
        shutil.copy2(path_sdist, dst_dir)

    archive_name = str(pathlib.Path(session.invoked_from, session.posargs[0]))
    make_archive(archive_name, root_dir=root_dir, base_dir=base_dir)


@nox.session
def publish(session: nox.Session) -> None:
    session.install("twine >= 6.0.1, < 7")

    nox_session_install_only_end(session)

    if not session.posargs:
        session.error("Path to API token file was not provided")

    dist = pathlib.Path("dist")
    dist_files = [path for path in dist.iterdir() if path.is_file()]
    dist_args = [str(path) for path in dist_files]

    session.run("twine", "check", "--strict", *dist_args)

    upload_args: list[str] = []
    upload_args.append("--non-interactive")
    upload_args.append("--disable-progress-bar")
    upload_args.extend(dist_args)

    env = session.env.copy()
    env["TWINE_USERNAME"] = "__token__"
    env["TWINE_PASSWORD"] = (
        pathlib.Path(session.invoked_from, session.posargs[0]).read_text().strip()
    )

    nox_command_run(
        ["twine", "upload", *upload_args],
        env=env,
        paths=session.bin_paths,
    )


def get_test_classes_from_env() -> Optional[Sequence[str]]:
    run_tests = os.environ.get("RUN_TESTS", "1")
    if not run_tests:
        return []

    try:
        if not int(run_tests):
            return ["InfoTest"]

        return None
    except ValueError:
        return run_tests.split(",")


@nox.session
def test_run(session: nox.Session) -> None:
    session.install("-e", ".")

    nox_session_install_only_end(session)

    opts: dict[str, object] = {
        opt: None for opt in ["always_write", "no_write", "readback", "no_readback"]
    }
    samples: Optional[pathlib.Path] = None
    test_classes: list[str] = []

    for index, arg in enumerate(session.posargs):
        if "--" not in session.posargs[:index] and arg.startswith("-"):
            if not arg.startswith("--"):
                session.error("Short options are not supported")

            if arg != "--":
                opts[arg.removeprefix("--").replace("-", "_")] = ...
        elif not samples:
            samples = pathlib.Path(session.invoked_from, arg)
        else:
            test_classes.append(arg)

    if not samples:
        session.error("Path to test samples was not provided")

    executable = ["no_vtf"]
    runner = TestRunner(cmd=executable, env=session.env, paths=session.bin_paths)
    test_suite = TestSuite(
        test_classes=(test_classes or None), runner=runner, samples=samples, session=session
    )

    test_suite.run([], opts)


@nox.session(reuse_venv=True)
def test_write(session: nox.Session) -> None:
    nox_session_install_only_end(session)

    session.notify("test_run", ["--always-write", "--readback", "--", *session.posargs])


@nox.session(reuse_venv=True)
def test_readback(session: nox.Session) -> None:
    nox_session_install_only_end(session)

    session.notify("test_run", ["--no-write", "--readback", "--", *session.posargs])


def test_update(session: nox.Session) -> None:
    nox_session_install_only_end(session)

    if not session.posargs:
        session.error("Path to new test samples was not provided")

    new_dir = f"{str(pathlib.Path(session.invoked_from, session.posargs[0])).rstrip('/')}/"

    old_dir: Union[pathlib.Path, str]
    with SampleArchives().extract() as old_dir:
        old_dir = f"{str(old_dir).rstrip('/')}/"

        with tempfile.TemporaryDirectory() as diff_dir:
            diff_dir = f"{diff_dir.rstrip('/')}/"

            session.run(
                *[
                    "rsync",
                    "--secluded-args",
                    *(3 * ["--human-readable"]),
                    *["--out-format", "+ %n"],
                    *["--archive", "--no-times"],
                    "--checksum",  # --times caused unnecessary transfers, use the next safest alt.
                    "--prune-empty-dirs",
                    *["--compare-dest", old_dir],
                    "--",
                    *[new_dir, diff_dir],
                ],
                external=True,
            )

            for dirpath, _, filenames in os_walk_strict(old_dir):
                for filename in filenames:
                    old_file = pathlib.Path(dirpath, filename)
                    old_file_relative = old_file.relative_to(old_dir)
                    new_file = pathlib.Path(new_dir, old_file_relative)

                    if not new_file.exists():
                        print(f"- {old_file_relative}")

                        removed_marker_file = pathlib.Path(
                            diff_dir,
                            old_file_relative.parts[0],
                            SampleArchives.REMOVED_SUBDIR,
                            *old_file_relative.parts[1:],
                        )
                        removed_marker_file.parent.mkdir(parents=True, exist_ok=True)
                        removed_marker_file.touch(exist_ok=False)

            session.run(
                *["find", diff_dir],
                *["-maxdepth", "1"],
                *["-type", "f"],
                *["-exec", "rm", "--", "{}", "+"],
                *["-printf", "- %P\\n"],
                external=True,
            )
            session.run(
                *["find", diff_dir],
                *["-type", "d"],
                "-empty",
                "-delete",
                external=True,
            )

            if not pathlib.Path(diff_dir).exists():
                session.skip()

            sample_archives = SampleArchives()
            for path in pathlib.Path(diff_dir).iterdir():
                assert path.is_dir()  # noqa S101 debug check

                new_archive = sample_archives.allocate(
                    path.relative_to(path.parent), ".tar.xz"
                ).resolve(strict=True)

                with session.chdir(path):
                    session.run(
                        "tar",
                        *["--create", "--xz"],
                        *["--file", str(new_archive)],
                        "--",
                        ".",
                        env={"XZ_OPT": "-9"},
                        external=True,
                    )


if os.name == "posix":
    test_update = nox.session(reuse_venv=True)(test_update)


@nox.session(reuse_venv=True)
def test_extract(session: nox.Session) -> None:
    nox_session_install_only_end(session)

    if not session.posargs:
        session.error("Destination path was not provided")
    destination = pathlib.Path(session.invoked_from, session.posargs[0])

    with SampleArchives().extract() as samples:
        shutil.copytree(
            samples, destination, symlinks=True, copy_function=shutil.copy2, dirs_exist_ok=True
        )


def nox_session_is_install_only(session: nox.Session) -> bool:
    logger = nox.logger.logger
    logger_level = logger.level
    try:
        logger.setLevel(logging.WARNING)
        return session.run("python", "--version", silent=True, log=False) is None
    finally:
        logger.setLevel(logger_level)


def nox_session_install_only_end(session: nox.Session) -> None:
    if nox_session_is_install_only(session):
        session.log("Skipping rest of the session, as --install-only is set.")
        session.skip()


class NoxSessionRunner(Protocol):
    def __call__(
        self,
        *args: str | os.PathLike[str],
        env: Mapping[str, str | None] | None = None,
        include_outer_env: bool = True,
        silent: bool = False,
        success_codes: Iterable[int] | None = None,
        log: bool = True,
        external: nox.command.ExternalType | None = None,
        stdout: int | IO[str] | None = None,
        stderr: int | IO[str] = subprocess.STDOUT,
        interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
        terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
    ) -> Any | None: ...


def nox_session_get_runner(
    session: nox.Session, *, install_only: bool = False, no_install: bool = True
) -> NoxSessionRunner:
    match nox_session_is_install_only(session), install_only, no_install:
        case _, False, True:
            return session.run
        case _, True, False:
            return session.run_install
        case False, True, True:
            return session.run
        case True, True, True:
            return session.run_install
        case False, False, False:
            return session.run_install
        case True, False, False:
            return session.run

    assert False, "unreachable"  # noqa S101


@overload
def nox_session_run_pip(
    runner: NoxSessionRunner,
    *args: str,
    env: Mapping[str, str] | None = None,
    include_outer_env: bool = True,
    silent: Literal[False] = False,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> Optional[bool]: ...


@overload
def nox_session_run_pip(
    runner: NoxSessionRunner,
    *args: str,
    env: Mapping[str, str] | None = None,
    include_outer_env: bool = True,
    silent: Literal[True] = True,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> Optional[str]: ...


@overload
def nox_session_run_pip(
    runner: NoxSessionRunner,
    *args: str,
    env: Mapping[str, str] | None = None,
    include_outer_env: bool = True,
    silent: bool = False,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> Optional[str | bool]: ...


def nox_session_run_pip(
    runner: NoxSessionRunner,
    *args: str,
    env: Mapping[str, str] | None = None,
    include_outer_env: bool = True,
    silent: bool = False,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> Optional[str | bool]:
    session = getattr(runner, "__self__", None)
    if not isinstance(session, nox.Session):
        raise ValueError(f"runner.__self__: expected nox.Session, got {type(session).__name__}")

    if external is None:
        external = "error" if nox.options.error_on_external_run else False

    env = {
        **(env or {}),
        "PIP_DISABLE_PIP_VERSION_CHECK": "1",
        "PIP_NO_INPUT": "1",
        "PIP_PROGRESS_BAR": "off",
        "PIP_NO_PYTHON_VERSION_WARNING": "1",
    }

    pip = (
        ["uv", "--no-progress", "pip"] if session.venv_backend == "uv" else ["python", "-m", "pip"]
    )

    return runner(
        *pip,
        *args,
        env=env,
        include_outer_env=include_outer_env,
        silent=silent,
        success_codes=success_codes,
        log=log,
        external=external,
        stdout=stdout,
        stderr=stderr,
        interrupt_timeout=interrupt_timeout,
        terminate_timeout=terminate_timeout,
    )


@overload
def nox_command_run(
    args: Sequence[str | os.PathLike[str]],
    *,
    env: Mapping[str, str | None] | None = None,
    silent: Literal[False] = False,
    paths: Sequence[str] | None = None,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> bool: ...


@overload
def nox_command_run(
    args: Sequence[str | os.PathLike[str]],
    *,
    env: Mapping[str, str | None] | None = None,
    silent: Literal[True] = True,
    paths: Sequence[str] | None = None,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> str: ...


@overload
def nox_command_run(
    args: Sequence[str | os.PathLike[str]],
    *,
    env: Mapping[str, str | None] | None = None,
    silent: bool = False,
    paths: Sequence[str] | None = None,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> str | bool: ...


def nox_command_run(
    args: Sequence[str | os.PathLike[str]],
    *,
    env: Mapping[str, str | None] | None = None,
    silent: bool = False,
    paths: Sequence[str] | None = None,
    success_codes: Iterable[int] | None = None,
    log: bool = True,
    external: Optional[nox.command.ExternalType] = None,
    stdout: int | IO[str] | None = None,
    stderr: int | IO[str] = subprocess.STDOUT,
    interrupt_timeout: float | None = nox.popen.DEFAULT_INTERRUPT_TIMEOUT,
    terminate_timeout: float | None = nox.popen.DEFAULT_TERMINATE_TIMEOUT,
) -> str | bool:
    if external is None:
        external = "error" if nox.options.error_on_external_run else False
    assert external is not None  # noqa S101 type narrowing

    return nox.command.run(
        args,
        env=env,
        silent=silent,
        paths=paths,
        success_codes=success_codes,
        log=log,
        external=external,
        stdout=stdout,
        stderr=stderr,
        interrupt_timeout=interrupt_timeout,
        terminate_timeout=terminate_timeout,
    )


class _NoxCommandRunKwArgs(TypedDict, total=False):
    stdout: Union[int, IO[str]]
    stderr: Union[int, IO[str]]
    success_codes: Iterable[int]


class _TestRunnerCallKwArgs(TypedDict, total=False):
    additional_env: dict[str, str]


class _RunKwArgs(_NoxCommandRunKwArgs, _TestRunnerCallKwArgs):
    pass


@dataclass(frozen=True, kw_only=True)
class TestRunner:
    cmd: Sequence[str]
    env: Optional[dict[str, str]] = None
    paths: Optional[list[str]] = None
    external: Optional[Union[Literal["error"], bool]] = None

    def __call__(
        self,
        *args: str,
        **kwargs: Unpack[_RunKwArgs],
    ) -> Union[str, bool]:
        full_cmd = list(self.cmd) + list(args)

        env = {**(self.env or os.environ), **kwargs.get("additional_env", {})}

        nox_kwargs = _NoxCommandRunKwArgs()
        if (stdout := kwargs.get("stdout", None)) is not None:
            nox_kwargs["stdout"] = stdout
        if (stderr := kwargs.get("stderr", None)) is not None:
            nox_kwargs["stderr"] = stderr
        if (success_codes := kwargs.get("success_codes", None)) is not None:
            nox_kwargs["success_codes"] = success_codes

        return nox_command_run(
            full_cmd, env=env, paths=self.paths, external=self.external, **nox_kwargs
        )


@dataclass(frozen=True, kw_only=True)
class CoverageTestRunner(TestRunner):
    coverage_run: Sequence[str]

    cmd: Sequence[str] = (
        "timeout",
        "--verbose",
        "--kill-after=10s",
        "5s",
        "python",
        "-m",
        "coverage",
        "run",
    )
    external: Optional[Union[Literal["error"], bool]] = True

    def __call__(
        self,
        *args: str,
        **kwargs: Unpack[_RunKwArgs],
    ) -> Union[str, bool]:
        args = (*self.coverage_run, *args)

        # registering a SIGTERM signal handler may cause the process to hang
        # (https://coverage.readthedocs.io/en/latest/config.html#run-sigterm)
        # this workaround is used to ensure eventual progress in case of hang
        for i in itertools.count(start=1):
            try:
                return super().__call__(*args, **kwargs)
            except nox.command.CommandFailed as exception:
                if exception.reason != "Returned code 124" or i == 10:
                    raise exception

        raise AssertionError("iterating over itertools.count() does not end")


@dataclass(kw_only=True)
class Test:
    runner: TestRunner
    samples: pathlib.Path

    @classmethod
    @contextmanager
    def with_archived_samples(cls, *, runner: TestRunner, **kwargs: object) -> Iterator[Self]:
        with SampleArchives().extract() as samples:
            test = cls(runner=runner, samples=samples, **kwargs)
            yield test

    def write(
        self, args: Optional[Sequence[object]] = None, opts: Optional[dict[str, object]] = None
    ) -> None:
        args = args or []

        opts = opts or {}
        opts = {"always_write": ..., "no_readback": ..., "no_write": None, "readback": None} | opts

        self.run(args, opts)

    def readback(
        self, args: Optional[Sequence[object]] = None, opts: Optional[dict[str, object]] = None
    ) -> None:
        args = args or []

        opts = opts or {}
        opts = {"no_write": ..., "readback": ..., "always_write": None, "no_readback": None} | opts

        self.run(args, opts)

    @final
    def run(self, args: Sequence[object], opts: dict[str, object], /) -> None:
        if not self._pre_run(args, opts):
            return

        exception: Optional[Exception] = None
        try:
            self._run(args, opts)
        except Exception as ex:
            exception = ex

        self._post_run(exception)

    def _pre_run(self, args: Sequence[object], opts: dict[str, object]) -> bool:
        return True

    @abstractmethod
    def _run(self, args: Sequence[object], opts: dict[str, object], /) -> None: ...

    def _post_run(self, exception: Optional[Exception]) -> None:
        if exception:
            raise exception


_TC = TypeVar("_TC", bound="TestCase")
_PS = ParamSpec("_PS")
_TestCaseRun: TypeAlias = Callable[Concatenate[_TC, Sequence[object], dict[str, object], _PS], None]


@dataclass(kw_only=True)
class TestCase(Test):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath(type(self).__name__)

    def get_input_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("input")

    def get_output_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("output")

    def get_test_samples(self) -> pathlib.Path:
        return self.samples / self.get_samples_subdir()

    def get_input_path(self, subpath: StrPath) -> pathlib.Path:
        return self.get_test_samples() / self.get_input_subdir() / subpath

    def get_output_path(self, subpath: StrPath) -> pathlib.Path:
        return self.get_test_samples() / self.get_output_subdir() / subpath

    @staticmethod
    def and_(
        *decorators: Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                nonlocal run
                for decorator in decorators:
                    run = decorator(run)

                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def or_(
        *decorators: Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                for decorator in decorators:
                    decorator(run)(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def with_common_options(
        compress: Optional[bool] = False,
        no_progress: bool = True,
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                compress_value = ... if compress is True else None
                no_compress_value = ... if compress is False else None
                no_progress_value = ... if no_progress else None

                opts = {
                    "compress": compress_value,
                    "no_compress": no_compress_value,
                    "no_progress": no_progress_value,
                } | opts
                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def rw_mode(
        write: Optional[bool] = False,
        readback: bool = False,
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                if write is True:
                    opts = {"always_write": ..., "no_write": None} | opts
                elif write is False:
                    opts = {"no_write": ..., "always_write": None} | opts

                if readback:
                    opts = {"readback": ..., "no_readback": None} | opts
                else:
                    opts = {"no_readback": ..., "readback": None} | opts

                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def with_inputs(
        subpaths: Optional[Sequence[StrPath]] = None,
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                nonlocal subpaths
                subpaths = subpaths or [pathlib.PurePath()]
                input_paths = [self.get_input_path(subpath) for subpath in subpaths]

                args = input_paths + list(args)
                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def in_place() -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                opts = {"output_dir": None, "output_file": None} | opts
                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def to_directory(
        subpath: Optional[StrPath] = None,
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                nonlocal subpath
                subpath = subpath or pathlib.PurePath()
                output_path = self.get_output_path(subpath)

                opts = {"output_dir": output_path, "output_file": None} | opts
                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def to_file(
        subpath: StrPath,
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                output_path = self.get_output_path(subpath)

                opts = {"output_file": output_path, "output_dir": None} | opts
                run(self, args, opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def parametrize(
        *opt_sets: dict[str, object]
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                for opt_set in opt_sets:
                    parametrized_opts = {**opt_set} | opts
                    run(self, args, parametrized_opts, *_args, **_kwargs)

            return inner

        return decorator

    @staticmethod
    def test_filter(  # noqa: C901
        option: str,
        subpath: StrPath,
        num_items: int,
        *,
        animate: Optional[bool] = None,
        ldr_format: Optional[str] = None,
        hdr_format: Optional[str] = "skip",
    ) -> Callable[[_TestCaseRun[_TC, _PS]], _TestCaseRun[_TC, _PS]]:
        def decorator(run: _TestCaseRun[_TC, _PS]) -> _TestCaseRun[_TC, _PS]:
            def inner(
                self: _TC,
                args: Sequence[object],
                opts: dict[str, object],
                /,
                *_args: _PS.args,
                **_kwargs: _PS.kwargs,
            ) -> None:
                input_file = self.get_input_path(subpath)
                args = [input_file] + list(args)

                nonlocal ldr_format
                ldr_format = ldr_format or ("apng" if animate else "tiff")
                opts = {"ldr_format": ldr_format} | opts

                if hdr_format:
                    opts = {"hdr_format": hdr_format} | opts

                if animate is True:
                    opts = {"animate": ...} | opts
                elif animate is False:
                    opts = {"no_animate": ...} | opts

                indices_pos = [1, num_items - 1]
                indices_neg = [-index for index in indices_pos]
                indices = [None] + indices_pos + indices_neg
                steps = [None, -2]

                def iter_run(filter_str: str) -> None:
                    iter_opts = {option: filter_str} | opts

                    iter_subpath = pathlib.Path(subpath)
                    filter_path = filter_str.replace(":", ",")
                    output_stem = f"{iter_subpath.stem}_{filter_path}"

                    if animate:
                        output_subpath = iter_subpath.with_name(f"{output_stem}.{ldr_format}")
                        output_file = self.get_output_path(output_subpath)

                        iter_opts = {"output_file": output_file} | iter_opts
                    else:
                        output_dir = self.get_output_path(output_stem)
                        output_dir.mkdir(parents=True, exist_ok=True)

                        iter_opts = {"output_dir": output_dir} | iter_opts

                    run(self, args, iter_opts, *_args, **_kwargs)

                for index in indices:
                    index = index or 0
                    if index < -num_items or index >= num_items:
                        continue

                    iter_run(str(index))

                for start, stop, step in itertools.product(indices, indices, steps):
                    if not len(range(num_items)[slice(start, stop, step)]):
                        continue

                    def str_optional(o: Optional[object], /) -> str:
                        return str(o) if o is not None else ""

                    filter_str = ""
                    filter_str += f"{str_optional(start)}:{str_optional(stop)}"
                    if step is not None:
                        filter_str += f":{str_optional(step)}"

                    iter_run(filter_str)

            return inner

        return decorator

    @with_common_options()
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        if opts.get("compress") and opts.get("no_compress"):
            raise ValueError("compress and no_compress are exclusive")

        if opts.get("always_write") and opts.get("no_write"):
            raise ValueError("always_write and no_write are exclusive")

        if opts.get("readback") and opts.get("no_readback"):
            raise ValueError("readback and no_readback are exclusive")

        parameters = self._str_opts(opts) + self._str_args(args)
        self.runner(*parameters, **_kwargs)

    def _str_opts(self, opts: dict[str, object]) -> list[str]:
        str_opts: list[str] = []
        for name, value in opts.items():
            if value is None:
                continue

            str_opt = "--" + name.replace("_", "-")
            str_opts.append(str_opt)

            if value is not Ellipsis:
                value = str(value)
                str_opts.append(value)

        return str_opts

    def _str_args(self, args: Sequence[object]) -> list[str]:
        if not args:
            return []

        str_args = ["--"]
        str_args.extend(str(arg) for arg in args)
        return str_args


@dataclass(kw_only=True)
class TestSuite(Test):
    test_classes: Optional[Sequence[Union[type[TestCase], str]]] = None

    session: Optional[nox.Session] = None

    @overload
    @classmethod
    def register(
        cls, test_class: type[_TC], /, *, condition: Optional[bool] = None
    ) -> type[_TC]: ...

    @overload
    @classmethod
    def register(
        cls, /, *, condition: Optional[bool] = None
    ) -> Callable[[type[_TC]], type[_TC]]: ...

    @classmethod
    def register(
        cls, test_class: Optional[type[_TC]] = None, /, *, condition: Optional[bool] = None
    ) -> Union[type[_TC], Callable[[type[_TC]], type[_TC]]]:
        if not test_class:
            return functools.partial(cls.register, condition=condition)

        if condition is not False:
            cls._registered_test_classes[test_class.__name__] = test_class

        return test_class

    _registered_test_classes: ClassVar[dict[str, type[TestCase]]] = {}

    def __post_init__(self) -> None:
        self._test_instances: list[TestCase] = []

        test_classes = self.test_classes or self._registered_test_classes.values()
        for test_class in test_classes:
            if isinstance(test_class, str):
                test_class = self._registered_test_classes[test_class]

            test_instance = test_class(runner=self.runner, samples=self.samples)
            self._test_instances.append(test_instance)

    def _run(self, args: Sequence[object], opts: dict[str, object], /) -> None:
        for test_instance in self._test_instances:
            if self.session:
                self.session.log(f"Running {type(test_instance).__name__}: {args = } {opts = }")

            test_instance.run(args, opts)


@TestSuite.register
@dataclass(kw_only=True)
class InfoTest(TestCase):
    @TestCase.or_(
        TestCase.parametrize({"help": ...}),  # pyright: ignore [reportArgumentType]
        TestCase.parametrize({"version": ...}),
        TestCase.parametrize({"credits": ...}),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class WriteSelfTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    def get_output_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("write_self_test_output")

    def _pre_run(self, args: Sequence[object], opts: dict[str, object]) -> bool:
        if "no_write" in opts and opts["no_write"] is not None:
            return False

        output_path = self.get_output_path("")
        shutil.rmtree(output_path, ignore_errors=True)  # pyright: ignore [reportDeprecated]
        output_path.mkdir()

        return True

    @TestCase.rw_mode(write=None)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)

    def _post_run(self, exception: Optional[Exception]) -> None:
        super()._post_run(exception)

        input_path = self.get_input_path("")
        output_path = self.get_output_path("")

        num_output_files_expected = 0
        for path in input_path.iterdir():
            if not path.is_file():
                continue

            stem_re = re.escape(path.stem)
            extension_re = r"\.[^\./]+$"

            envmap_6_faces_re = r"(?:bk|dn|ft|lf|rt|up)"
            envmap_7_faces_re = r"(?:" + envmap_6_faces_re + "|sph)"
            match path.stem:
                case "envmap_6_faces":
                    additional_re = envmap_6_faces_re
                case "envmap_7_faces":
                    additional_re = envmap_7_faces_re
                case _:
                    additional_re = r""

            pattern = re.compile(stem_re + additional_re + extension_re, re.ASCII)

            output_files = list(path_re_glob(output_path, pattern))
            num_output_files_expected += len(output_files)

        num_output_files_actual = len([path for path in output_path.iterdir() if path.is_file()])
        if num_output_files_actual != num_output_files_expected:
            raise ValueError("Number of expected and actual written files must match")

        shutil.rmtree(output_path)  # pyright: ignore [reportDeprecated]


@TestSuite.register
@dataclass(kw_only=True)
class ReadbackNegativeSelfTest(TestCase):
    def get_input_path(self, subpath: StrPath) -> pathlib.Path:
        return self.samples / subpath

    def get_output_path(self, subpath: StrPath) -> pathlib.Path:
        return self.samples / subpath

    def _pre_run(self, args: Sequence[object], opts: dict[str, object]) -> bool:
        return not any(
            (
                ("no_write" in opts and opts["no_write"] is None),
                ("readback" in opts and opts["readback"] is None),
                ("always_write" in opts and opts["always_write"] is not None),
                ("no_readback" in opts and opts["no_readback"] is not None),
            )
        )

    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(["FormatTest/input/rgb888.vtf"]),
            TestCase.to_directory("FormatTest/nonexistent_directory"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["FormatTest/input/rgb888.vtf"]),
            TestCase.to_file("FormatTest/output/nonexistent_file.tiff"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["FormatTest/input/rgb888.vtf"]),
            TestCase.to_file("FormatTest/output/rgba8888.tiff"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["FormatTest/input/bgra8888.vtf"]),
            TestCase.to_file("FormatTest/output/bgra8888_hdr.tiff"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["MultipleFramesTest/input/8x8_.vtf"]),
            TestCase.to_file("MultipleFramesTest/output/8x8__,-1.apng"),
            TestCase.parametrize({"animate": ..., "ldr_format": "apng"}),
        ),
        TestCase.and_(
            TestCase.with_inputs(["FormatTest/input/rgb888.vtf"]),
            TestCase.to_file("FormatTest/output/rgba8888.raw"),
            TestCase.parametrize({"raw": ...}),
        ),
        TestCase.and_(
            TestCase.with_inputs(["FormatTest/input/i8.vtf"]),
            TestCase.to_file("FormatTest/output/i8_extra_null_byte.raw"),
            TestCase.parametrize({"raw": ...}),
        ),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        _kwargs = {"success_codes": [1, 2], **_kwargs}
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class FormatTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.parametrize({}, {"mipmaps": ...})
    @TestCase.parametrize({}, {"separate_channels": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        ldr_formats = ["png", "tga", "tiff", "apng"]
        for ldr_format in ldr_formats:
            super()._run(args, {"ldr_format": ldr_format, "hdr_format": "skip"} | opts, **_kwargs)

        hdr_formats = ["exr", "tiff"]
        for hdr_format in hdr_formats:
            super()._run(args, {"ldr_format": "skip", "hdr_format": hdr_format} | opts, **_kwargs)

        super()._run(args, {"raw": ...} | opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class MultipleFramesTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(["8x8_.vtf"]),
            TestCase.to_directory(),
            TestCase.parametrize({"no_animate": ...}),
            TestCase.parametrize({}, {"mipmaps": ...}),
            TestCase.parametrize({}, {"separate_channels": ...}),
        ),
        TestCase.test_filter("frames", "8x8_.vtf", 64, animate=False),
        TestCase.test_filter("frames", "8x8_.vtf", 64, animate=True),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class MultipleSlicesTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(["volume.vtf"]),
            TestCase.to_directory(),
            TestCase.parametrize({}, {"mipmaps": ...}),
            TestCase.parametrize({}, {"separate_channels": ...}),
        ),
        TestCase.test_filter("slices", "volume.vtf", 256),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class MultipleFacesTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.test_filter("faces", "envmap_7_faces.vtf", 7)
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class VersionTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.with_inputs(["rgb888_v7.0.vtf"]),  # pyright: ignore [reportArgumentType]
        TestCase.with_inputs(["rgb888_v7.1.vtf"]),
        TestCase.with_inputs(["rgb888_v7.2.vtf"]),
        TestCase.with_inputs(["rgb888_v7.3.vtf"]),
        TestCase.with_inputs(["rgb888_v7.4.vtf"]),
        TestCase.with_inputs(["rgb888_v7.5.vtf"]),
    )
    @TestCase.to_directory()
    @TestCase.parametrize({}, {"mipmaps": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class LowResolutionImageTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("VersionTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.with_inputs(["rgb888_v7.0.vtf"]),  # pyright: ignore [reportArgumentType]
        TestCase.with_inputs(["rgb888_v7.1.vtf"]),
        TestCase.with_inputs(["rgb888_v7.2.vtf"]),
        TestCase.with_inputs(["rgb888_v7.3.vtf"]),
        TestCase.with_inputs(["rgb888_v7.4.vtf"]),
        TestCase.with_inputs(["rgb888_v7.5.vtf"]),
    )
    @TestCase.to_directory("low_res_img")
    @TestCase.parametrize({"low_res_img": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class PathTest(TestCase):
    def get_input_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath()

    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(
                [
                    "input/a.vtf/b/c.vtf",
                    "input/a.vtf",
                    "input/a.vtf/b",
                    "input",
                    "input/a.vtf/b/c.vtf",
                ]
            ),
            TestCase.to_directory("to_directory"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["input/a.vtf/b/c.vtf"]),
            TestCase.to_file("to_file/c.tiff"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["in_place"]),
            TestCase.in_place(),
        ),
    )
    @TestCase.parametrize({"ldr_format": "png", "hdr_format": "skip"})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class UnsupportedFormatTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.parametrize({"ldr_format": "ppm", "hdr_format": "skip", "separate_channels": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class SequentialRunnerTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.parametrize({"num_workers": 1})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class ResolutionFilterTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs(["rgb888.vtf"])
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.to_file("rgb888_min_512px.tiff"),
            TestCase.parametrize({"min_resolution": 512}),
        ),
        TestCase.and_(
            TestCase.to_file("rgb888_min_2048px_closest.tiff"),
            TestCase.parametrize({"min_resolution": 2048, "closest_resolution": ...}),
        ),
        TestCase.and_(
            TestCase.to_file("rgb888_max_256px.tiff"),
            TestCase.parametrize({"max_resolution": 256}),
        ),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class ProgressBarTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.with_common_options(no_progress=False)
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class HdrToLdrTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(
                [
                    "bgra8888_hdr.vtf",
                    "rgb0-rgba16161616f_hdr.vtf",
                    "rgba16161616f_hdr.vtf",
                    "rgba16161616_hdr.vtf",
                ]
            ),
            TestCase.to_directory("hdr_to_ldr"),
        ),
        TestCase.and_(
            TestCase.with_inputs(["bgra8888.vtf"]),
            TestCase.to_directory(),
        ),
    )
    @TestCase.parametrize({"hdr_to_ldr": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class RawIsUnaffectedByOtherFlagsTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.parametrize({"raw": ...})
    @TestCase.parametrize({"hdr_to_ldr": ..., "separate_channels": ...})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class PipelineStepsTest(TestCase):
    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs()
    @TestCase.to_directory()
    @TestCase.parametrize({"no_animate": ...}, {"ldr_format": "apng"})
    @TestCase.or_(
        TestCase.parametrize({}, {"raw": ...}),  # pyright: ignore [reportArgumentType]
        TestCase.and_(
            TestCase.parametrize({}, {"hdr_to_ldr": ...}),
            TestCase.parametrize({}, {"separate_channels": ...}),
        ),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class DynamicRangeOverrideTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.or_(
        TestCase.and_(  # pyright: ignore [reportArgumentType]
            TestCase.with_inputs(["bgra8888.vtf"]),
            TestCase.to_file("bgra8888_hdr.exr"),
            TestCase.parametrize({"dynamic_range": "hdr"}),
        ),
        TestCase.and_(
            TestCase.with_inputs(["bgra8888_hdr.vtf"]),
            TestCase.to_file("bgra8888.tiff"),
            TestCase.parametrize({"dynamic_range": "ldr"}),
        ),
    )
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class OverbrightFactorTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("FormatTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs(["bgra8888_hdr.vtf"])
    @TestCase.to_file("bgra8888_hdr_overbright_factor_8.exr")
    @TestCase.parametrize({"overbright_factor": 8})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


@TestSuite.register
@dataclass(kw_only=True)
class CommaSeparatedFramesFilterTest(TestCase):
    def get_samples_subdir(self) -> pathlib.PurePath:
        return pathlib.PurePath("MultipleFramesTest")

    @TestCase.rw_mode(readback=True)
    @TestCase.with_inputs(["8x8_.vtf"])
    @TestCase.to_file("8x8__,,9;-2,-8,-1;-8,1,-8.apng")
    @TestCase.parametrize({"frames": "::9,-2:-8:-1,-8:1:-8"})
    def _run(
        self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
    ) -> None:
        super()._run(args, opts, **_kwargs)


if sys.platform == "linux":  # noqa: C901
    _OutputStream: TypeAlias = Optional[pathlib.Path]
    _OutputStreams: TypeAlias = tuple[_OutputStream, _OutputStream]

    @TestSuite.register
    @dataclass(kw_only=True)
    class ShellCompletionTest(TestCase):
        @TestCase.rw_mode(readback=True)
        @TestCase.with_common_options(compress=None, no_progress=False)
        def _run(
            self, args: Sequence[object], opts: dict[str, object], /, **_kwargs: Unpack[_RunKwArgs]
        ) -> None:
            if opts.get("always_write") and opts.get("no_write"):
                raise ValueError("always_write and no_write are exclusive")

            if opts.get("readback") and opts.get("no_readback"):
                raise ValueError("readback and no_readback are exclusive")

            write = opts.pop("always_write", None) is not None
            readback = opts.pop("readback", None) is not None
            opts.pop("no_write", None)
            opts.pop("no_readback", None)

            unsupported_shell = "unsupported"

            def run(shell: str, actual: _OutputStreams, expected: _OutputStreams) -> None:
                stdout, stderr = actual

                opener = functools.partial(mkdir_parents_opener, mode=0o666)
                open_output = functools.partial(open, mode="wb", opener=opener)
                open_pty = contextmanager_bind(
                    dummy_pty(), functools.partial(os.fdopen, mode="wb", closefd=False)
                )
                with optional_contexts(
                    open_output(stdout) if stdout else open_pty,
                    open_output(stderr) if stderr else None,
                ) as (stdout_file, stderr_file):
                    run_kwargs = _RunKwArgs(
                        additional_env={"_CONSOLE_SCRIPT_PATH_OVERRIDE": "/test/path/no_vtf"}
                    )

                    if stdout_file:
                        run_kwargs["stdout"] = stdout_file.fileno()

                    if stderr_file:
                        run_kwargs["stderr"] = stderr_file.fileno()

                    if stdout and shell == unsupported_shell:
                        run_kwargs["success_codes"] = [1]

                    run_kwargs.update(_kwargs)

                    super(ShellCompletionTest, self)._run(
                        args, {"shell_completion": shell} | opts, **run_kwargs
                    )

                for actual_stream, expected_stream in zip(actual, expected, strict=True):
                    if not (actual_stream and expected_stream):
                        continue

                    nox_command_run(
                        [
                            *["git", "--no-pager", "--literal-pathspecs"],
                            *["diff", "--no-index", "--exit-code", "--unified", "--histogram"],
                            *["--src-prefix", "expected/"],
                            *["--dst-prefix", "actual/"],
                            "--",
                            *[str(expected_stream), str(actual_stream)],
                        ],
                        env={
                            **os.environ,
                            "GIT_CONFIG_GLOBAL": "/dev/null",
                            "GIT_CONFIG_SYSTEM": "/dev/null",
                        },
                    )

            for shell, redirect_stdout in itertools.product(
                [unsupported_shell, "bash", "zsh", "fish"], [False, True]
            ):
                test_step_path = (
                    self.get_test_samples()
                    / shell
                    / ("stdout_file" if redirect_stdout else "stdout_tty")
                )
                expected_stdout: Optional[pathlib.Path] = test_step_path / "stdout"
                expected_stderr: Optional[pathlib.Path] = test_step_path / "stderr"

                with optional_contexts(
                    (expected_stdout if write else tempfile_name()) if redirect_stdout else None,
                    expected_stderr if write else tempfile_name(),
                    expected_stdout if readback and redirect_stdout else None,
                    expected_stderr if readback else None,
                ) as (actual_stdout, actual_stderr, expected_stdout, expected_stderr):
                    run(shell, (actual_stdout, actual_stderr), (expected_stdout, expected_stderr))


if sys.platform == "linux":

    @contextmanager
    def dummy_pty() -> Iterator[int]:
        def copy() -> None:
            with contextlib.suppress(OSError):
                while os.read(master_fd, 16 * 1024):
                    pass

        thread = threading.Thread(target=copy, daemon=True)

        master_fd, slave_fd = pty.openpty()
        try:
            fcntl.ioctl(slave_fd, termios.TIOCEXCL)
            tty.setraw(master_fd, termios.TCSANOW)

            thread.start()

            yield slave_fd
        finally:
            os.close(slave_fd)

            if thread.ident:
                thread.join()

            os.close(master_fd)


@contextmanager
def tempfile_name() -> Iterator[pathlib.Path]:
    with tempfile.NamedTemporaryFile() as named_tempfile:
        yield pathlib.Path(named_tempfile.name)


class SampleArchives:
    REMOVED_SUBDIR: Final = pathlib.PurePath("._removed")

    def __init__(self, top: Optional[pathlib.Path] = None) -> None:
        self.top: Final = top or pathlib.Path("resources/test/samples")

    def subdirs(self) -> Iterator[pathlib.PurePath]:
        for path in self.top.iterdir():
            if path.is_dir():
                yield path.relative_to(self.top)

    def archives(self, subdir: pathlib.PurePath) -> Sequence[pathlib.Path]:
        archives_dir = self._get_archives_dir(subdir)
        if not archives_dir.exists():
            return []

        return sorted(
            (
                path
                for path in archives_dir.iterdir()
                if not (path.is_dir() or path.name.endswith(".license"))
            ),
            key=lambda path: (len(path.name), path),
        )

    def allocate(self, subdir: pathlib.PurePath, extension: str) -> pathlib.Path:
        new_archive_stem = "1"
        if archives := self.archives(subdir):
            new_archive_stem = str(int(archives[-1].name.split(".")[0]) + 1)

        new_archive = (self._get_archives_dir(subdir) / new_archive_stem).with_suffix(extension)

        new_archive.parent.mkdir(parents=True, exist_ok=True)
        new_archive.touch(exist_ok=False)

        return new_archive

    @contextmanager
    def extract(self) -> Iterator[pathlib.Path]:
        with tempfile.TemporaryDirectory() as output_name:
            output_directory = pathlib.Path(output_name)

            for subdir in self.subdirs():
                extract_dir = output_directory / subdir
                extract_dir.mkdir(parents=True, exist_ok=True)

                for archive in self.archives(subdir):
                    shutil.unpack_archive(archive, extract_dir)

                    removed_dir = extract_dir / self.REMOVED_SUBDIR
                    if removed_dir.exists():
                        for dirpath, _, filenames in os_walk_strict(removed_dir):
                            for filename in filenames:
                                removed_marker = pathlib.Path(dirpath, filename)
                                removed_marker_relative = removed_marker.relative_to(removed_dir)

                                (extract_dir / removed_marker_relative).unlink(missing_ok=True)
                                (removed_dir / removed_marker_relative).unlink(missing_ok=False)

            for dirpath, dirnames, filenames in os_walk_strict(output_directory, topdown=False):
                # update dirnames: the list of subdirectories is retrieved before the tuples for the
                # directory and its subdirectories are generated; and the directories in dirnames
                # have already been generated by the time dirnames itself is generated
                dirnames = [
                    dirname for dirname in dirnames if pathlib.Path(dirpath, dirname).exists()
                ]

                if not (dirnames or filenames):
                    pathlib.Path(dirpath).rmdir()

            yield output_directory

    def _get_archives_dir(self, subdir: pathlib.PurePath) -> pathlib.Path:
        return pathlib.Path(self.top, subdir)


@contextmanager
def optional_contexts(
    *args: Optional[Union[_T_co, AbstractContextManager[_T_co]]],
) -> Iterator[tuple[Optional[_T_co], ...]]:
    with contextlib.ExitStack() as context_stack:
        yield tuple(
            (
                context_stack.enter_context(arg)  # pyright: ignore [reportUnknownArgumentType]
                if isinstance(arg, AbstractContextManager)
                else arg
            )
            for arg in args
        )


@contextmanager
def contextmanager_bind(
    cm_a: AbstractContextManager[_T_co],
    a_to_cm_b: Callable[[_T_co], AbstractContextManager[_T_contra]],
    /,
) -> Iterator[_T_contra]:
    with cm_a as a:
        with a_to_cm_b(a) as b:
            yield b


def os_walk_strict(
    top: StrPath, topdown: bool = True, followlinks: bool = False
) -> Iterator[tuple[str, list[str], list[str]]]:
    def onerror(exception: OSError) -> None:
        raise exception

    return os.walk(
        top=top,
        topdown=topdown,
        onerror=onerror,
        followlinks=followlinks,
    )


def mkdir_parents_opener(
    path: str, flags: int, mode: int = 0o777, *, dir_fd: Optional[int] = None
) -> int:
    if flags & (os.O_WRONLY | os.O_RDWR):
        pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)

    return os.open(path, flags, mode, dir_fd=dir_fd)


def make_archive(
    archive_name: str,
    *,
    root_dir: Optional[StrPath] = None,
    base_dir: Optional[StrPath] = None,
    dry_run: bool = False,
    owner: Optional[str] = None,
    group: Optional[str] = None,
    logger: object = None,
) -> str:
    archive_base_name: str
    archive_format: str
    if (archive_base_name := archive_name.removesuffix(".tar.xz")) != archive_name:
        archive_format = "xztar"
    elif (archive_base_name := archive_name.removesuffix(".zip")) != archive_name:
        archive_format = "zip"
    else:
        raise RuntimeError("Unsupported archive format")

    return shutil.make_archive(
        archive_base_name,
        archive_format,
        root_dir=root_dir,
        base_dir=base_dir,
        dry_run=dry_run,
        owner=owner,
        group=group,
        logger=logger,
    )


def path_re_glob(path: pathlib.Path, pattern: re.Pattern[str]) -> Iterable[pathlib.Path]:
    return (path for path in path.iterdir() if pattern.fullmatch(path.name))



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final, Optional

from no_vtf.core.image import ImageDynamicRange
from no_vtf.core.texture import Texture, TextureIndex
from no_vtf.typing import mypyc_attr

from .generated.parser import VtfImage as VtfParserImage


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfTexture(Texture):
    def __init__(
        self,
        *,
        dynamic_range: Optional[ImageDynamicRange],
        onebitalpha: Optional[bool] = None,
        eightbitalpha: Optional[bool] = None,
        width: int,
        height: int,
        mipmap: Optional[TextureIndex],
        frame: Optional[TextureIndex],
        face: Optional[TextureIndex],
        slice: Optional[TextureIndex],  # noqa: A002
        image: VtfParserImage,
    ) -> None:
        super().__init__(
            width=width,
            height=height,
            mipmap=mipmap,
            frame=frame,
            face=face,
            slice=slice,
        )

        self.dynamic_range: Final = dynamic_range
        self.onebitalpha: Final = onebitalpha
        self.eightbitalpha: Final = eightbitalpha
        self.image: Final = image
